﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.ComponentModel;
using System.Linq;
using System.Windows.Forms;
using HeuristicLab.Clients.Access;
using HeuristicLab.Core.Views;
using HeuristicLab.MainForm;
using Tasks = System.Threading.Tasks;

namespace HeuristicLab.Clients.Hive.Administrator.Views {
  [View("ProjectView")]
  [Content(typeof(Project), IsDefaultView = true)]
  public partial class ProjectView : ItemView {
    private readonly object locker = new object();
    private bool updatingUsers = false;

    public new Project Content {
      get { return (Project)base.Content; }
      set { base.Content = value; }
    }

    public ProjectView() {
      InitializeComponent();

      AccessClient.Instance.Refreshing += AccessClient_Instance_Refreshing;
      AccessClient.Instance.Refreshed += AccessClient_Instance_Refreshed;
    }

    protected override void OnClosing(FormClosingEventArgs e) {
      AccessClient.Instance.Refreshed -= AccessClient_Instance_Refreshed;
      AccessClient.Instance.Refreshing -= AccessClient_Instance_Refreshing;
      base.OnClosing(e);
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      Content.PropertyChanged += Content_PropertyChanged;
    }

    protected override void DeregisterContentEvents() {
      Content.PropertyChanged -= Content_PropertyChanged;
      base.DeregisterContentEvents();
    }

    private void Content_PropertyChanged(object sender, PropertyChangedEventArgs e) {
      OnContentChanged();
    }

    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content == null) {
        nameTextBox.Clear();
        descriptionTextBox.Clear();
        ownerComboBox.DataSource = null;
        ownerComboBox.SelectedItem = null;
        createdTextBox.Clear();
        startDateTimePicker.Value = DateTime.Now;
        endDateTimePicker.Value = startDateTimePicker.Value;
        indefiniteCheckBox.Checked = false;
      } else {
        nameTextBox.Text = Content.Name;
        descriptionTextBox.Text = Content.Description;
        ownerComboBox.DataSource = AccessClient.Instance.UsersAndGroups.OfType<LightweightUser>().ToList();
        ownerComboBox.SelectedItem = AccessClient.Instance.UsersAndGroups.FirstOrDefault(x => x.Id == Content.OwnerUserId);
        createdTextBox.Text = Content.DateCreated.ToString("ddd, dd.MM.yyyy, HH:mm:ss");
        startDateTimePicker.Value = Content.StartDate;
        if (indefiniteCheckBox.Checked = !Content.EndDate.HasValue) {
          endDateTimePicker.Value = startDateTimePicker.Value;
        } else {
          endDateTimePicker.Value = Content.EndDate.Value;
        }
      }
    }

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      bool enabled = Content != null;
      nameTextBox.Enabled = enabled;
      descriptionTextBox.Enabled = enabled;
      ownerComboBox.Enabled = enabled;
      createdTextBox.Enabled = enabled;
      startDateTimePicker.Enabled = enabled;
      endDateTimePicker.Enabled = enabled && Content.EndDate.HasValue;
      indefiniteCheckBox.Enabled = enabled;
    }

    private void AccessClient_Instance_Refreshing(object sender, EventArgs e) {
      if (InvokeRequired) Invoke((Action<object, EventArgs>)AccessClient_Instance_Refreshing, sender, e);
      else {
        var mainForm = MainFormManager.GetMainForm<MainForm.WindowsForms.MainForm>();
        mainForm.AddOperationProgressToView(this, "Refreshing ...");
        SetEnabledStateOfControls();
      }
    }

    private void AccessClient_Instance_Refreshed(object sender, EventArgs e) {
      if (InvokeRequired) Invoke((Action<object, EventArgs>)AccessClient_Instance_Refreshed, sender, e);
      else {
        var mainForm = MainFormManager.GetMainForm<MainForm.WindowsForms.MainForm>();
        mainForm.RemoveOperationProgressFromView(this);
        SetEnabledStateOfControls();
      }
    }

    private async void ProjectView_Load(object sender, EventArgs e) {
      lock (locker) {
        if (updatingUsers) return;
        updatingUsers = true;
      }
      try {
        await Tasks.Task.Run(() => UpdateUsers());
      } finally {
        updatingUsers = false;
      }
    }

    private async void refreshButton_Click(object sender, EventArgs e) {
      lock (locker) {
        if (updatingUsers) return;
        updatingUsers = true;
      }
      try {
        await Tasks.Task.Run(() => UpdateUsers());
      } finally {
        updatingUsers = false;
      }
    }

    private void UpdateUsers() {
      ownerComboBox.DataSource = null;
      AccessClient.Instance.Refresh();
      if (Content != null) {
        ownerComboBox.DataSource = AccessClient.Instance.UsersAndGroups.OfType<LightweightUser>().ToList();
        ownerComboBox.SelectedItem = AccessClient.Instance.UsersAndGroups.FirstOrDefault(x => x.Id == Content.OwnerUserId);
      }
    }

    private void nameTextBox_Validating(object sender, CancelEventArgs e) {
      if (string.IsNullOrEmpty(nameTextBox.Text)) {
        MessageBox.Show(
          "Project must have a name.",
          "HeuristicLab Hive Administrator",
          MessageBoxButtons.OK,
          MessageBoxIcon.Error);
        e.Cancel = true;
      } else Content.Name = nameTextBox.Text;
    }

    private void descriptionTextBox_Validating(object sender, CancelEventArgs e) {
      Content.Description = descriptionTextBox.Text;
    }

    private void ownerComboBox_SelectedIndexChanged(object sender, EventArgs e) {
      var selectedItem = (LightweightUser)ownerComboBox.SelectedItem;
      Content.OwnerUserId = selectedItem != null ? selectedItem.Id : Guid.Empty;
    }

    private void startDateTimePicker_Validating(object sender, CancelEventArgs e) {
      if (!Content.EndDate.HasValue || startDateTimePicker.Value > Content.EndDate)
        endDateTimePicker.Value = startDateTimePicker.Value;
      Content.StartDate = startDateTimePicker.Value;
    }

    private void endDateTimePicker_Validating(object sender, CancelEventArgs e) {
      if (endDateTimePicker.Value < startDateTimePicker.Value)
        endDateTimePicker.Value = startDateTimePicker.Value;
      Content.EndDate = endDateTimePicker.Value;
    }

    private void indefiniteCheckBox_CheckedChanged(object sender, EventArgs e) {
      Content.EndDate = indefiniteCheckBox.Checked ? (DateTime?)null : Content.StartDate;
    }
  }
}
